# -------------------------------------------------------------------
# Plot samples.
# Double-couple (DC) solutions
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

from seishmc.utils.find_solutions import find_sol

import seaborn as sns
sns.set_theme(style="ticks", palette=None)

import matplotlib.pyplot as plt

import pandas as pd
import numpy as np
import pickle


def read_samples_DC(file_path):
    '''Read samples from pickle file and get the solution'''
    with open(file_path, 'rb') as f:
        samples = pickle.load(f)
        misfit = pickle.load(f)

    samples = np.asarray(samples)
    df = pd.DataFrame({
        'strike': samples[:, 0],
        'dip':    samples[:, 1],
        'rake':   samples[:, 2],
        'Mw':     np.round(samples[:, 3], 3),
        'misfit': misfit,
    })

    return df, find_sol(df)


def pairplot_samples_DC(file_path, fig_saving_path=None, init_sol=None, ref_sol=None):
    '''Plot samples in seaborn pairplot (FMT solutions)'''

    df, sol = read_samples_DC(file_path)
    make_figure_DC(df, fig_saving_path=fig_saving_path, init_sol=init_sol, ref_sol=ref_sol)


def make_figure_DC(df, fig_saving_path=None, init_sol=None, nq=4, ref_sol=None):
    '''Make figure based on the dataframe'''

    # determine the solution
    sol = find_sol(df)
    n_sample = len(df)

    # Using Seaborn
    g = sns.PairGrid(df, hue='misfit', corner=True)
    g.map_lower(sns.scatterplot, palette='viridis', marker=".", linewidth=0.5, legend=True)
    g.map_lower(sns.kdeplot, levels=6, color="lightgrey", thresh=.2, alpha=0.75, linewidths=0.5, hue=None)
    g.map_diag(sns.histplot,  color='.15', hue=None)

    for i in range(nq):
        for j in range(nq):
            if i <= j:
                continue
            else:
                # other solution (true in syn. or inverted with MTUQ).
                if ref_sol is not None:
                    g.axes[i, j].scatter(ref_sol[j], ref_sol[i], marker='^', s=200, c='orange', edgecolors='black', label='True')

                # inverted solution
                g.axes[i, j].scatter(sol[j], sol[i], marker='*', s=300, c='C3', edgecolors='white')

                # initial solution (q0)
                if init_sol is not None:
                    g.axes[i, j].scatter(init_sol[j], init_sol[i], marker='o', s=100, c='white', edgecolors='black', label='Initial')

    # set x-axis and ticks
    g.axes[nq - 1, 0].set_xlim([-20, 380])
    g.axes[nq - 1, 0].set_xticks([0, 180, 360])
    g.axes[nq - 1, 0].set_xticklabels([0, 180, 360])
    g.axes[nq - 1, 1].set_xlim([-20, 110])
    g.axes[nq - 1, 1].set_xticks([0, 45, 90])
    g.axes[nq - 1, 1].set_xticklabels([0, 45, 90])
    g.axes[nq - 1, 2].set_xlim([-110, 110])
    g.axes[nq - 1, 2].set_xticks([-90, 0, 90])
    g.axes[nq - 1, 2].set_xticklabels([-90, 0, 90])

    # set y-axis and ticks
    g.axes[1, 0].set_ylim([-20, 110])
    g.axes[1, 0].set_yticks([0, 45, 90])
    g.axes[1, 0].set_yticklabels([0, 45, 90])
    g.axes[2, 0].set_ylim([-110, 110])
    g.axes[2, 0].set_yticks([-90, 0, 90])
    g.axes[2, 0].set_yticklabels([-90, 0, 90])

    # set figure size and layout
    plt.gcf().set_size_inches(12, 10)
    plt.tight_layout()

    # save figures to file
    if fig_saving_path is not None:
        fig_formats = ['png', ]
        for fmt in fig_formats:
            plt.savefig('%s_PairView_N%d.%s' % (fig_saving_path, n_sample, fmt))
    else:
        plt.show()


# -------------------------------------------------------------------
# Moment tensor inversion using Hamiltonian Monte Carlo sampling
# Full moment tensor (FMT) solutions
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

from seishmc.utils import wrap_strike, wrap_rake, wrap_dip, wrap_colatitude, wrap_lunelongitude
from seishmc.DHMC.dc import DHMC_DC

from mtuq.grid import UnstructuredGrid
from mtuq.util.math import to_rho, to_v, to_w, to_mij
from mtuq.grid.moment_tensor import to_mt
from mtuq.event import MomentTensor

import numpy as np

class DHMC_FMT(DHMC_DC):
    ''' Utilizing the Hamiltonian Monte Carlo sampling method to sample the full moment tensor solutions.'''

    def __init__(self, misfit_bw, data_bw, greens_bw,
                 misfit_sw, data_sw, greens_sw,
                 saving_dir, b_save_cache=True, n_step_cache=100,
                 verbose=False):

        super(DHMC_FMT, self).__init__(misfit_bw, data_bw, greens_bw,
                                       misfit_sw, data_sw, greens_sw,
                                       saving_dir, b_save_cache, n_step_cache,
                                       verbose)

        # full moment tensor (FMT) solutions
        # [strike, dip, rake, magnitude, co-latitude, longitude]
        self.Nq = 6
        # mass matrix
        self.massinv = np.eye(self.Nq)
        self.massinv[3, 3] = 1.0 / 1000.     # magnitude
        # delta to calculate derivatives.
        self.delta_q0 = 0.2 * np.ones(self.Nq)
        self.delta_q0[3] = 0.02  # magnitude
        self.q0 = np.zeros(self.Nq)
        self.raw_q0 = np.zeros(self.Nq)
        self.dU_dqi = np.zeros(self.Nq)
        self.sol_misfit = np.zeros(self.Nq)

        self.calc_scaling_factor()

    def get_solution(self):
        '''Get solution'''

        sol = self._get_solution()
        kappa   = sol[0]  # strike
        h       = np.cos(np.deg2rad(sol[1]))  # dip
        sigma   = sol[2]  # rake
        rho     = to_rho(sol[3])  # magnitude: Mw
        v       = to_v(sol[5])
        w       = to_w(90. - sol[4])
        mij     = to_mij(rho, v, w, kappa, sigma, h)

        # lune dict
        self.lune_dict = dict({'rho': rho,
                          'v': v,
                          'w': w,
                          'kappa': kappa,
                          'sigma': sigma,
                          'h': h, })
        self.best_mt = MomentTensor(mij)
        return self.best_mt, self.lune_dict


    def Uq(self, q):
        '''Compute Uq and gradient'''

        npts = self.Nq + 1
        kappa = q[0] * np.ones(npts)                    # strike
        h = np.cos(np.deg2rad(q[1])) * np.ones(npts)    # dip
        sigma = q[2] * np.ones(npts)                    # rake
        rho = to_rho(q[3]) * np.ones(npts)              # magnitude: Mw
        w = to_w(90. - q[4]) * np.ones(npts)            # co-latitude to w
        v = to_v(q[5]) * np.ones(npts)                  # lune. longitude to v

        # x+dx
        kappa[1] += self.delta_q0[0]
        h[2] = np.cos(np.deg2rad(q[1] + self.delta_q0[1]))
        sigma[3] += self.delta_q0[2]
        rho[4] = to_rho(q[3] + self.delta_q0[3])
        w[5] = to_w(90. - (q[4] + self.delta_q0[4]))
        v[6] = to_v(q[5] + self.delta_q0[5])

        q_grid = UnstructuredGrid(
            dims=('rho', 'v', 'w', 'kappa', 'sigma', 'h'),
            coords=(rho, v, w, kappa, sigma, h),
            callback=to_mt)

        misfit = np.zeros([npts, 1])

        try:
            if self.data_bw is not None and self.greens_bw is not None:
                res_bw = self.misfit_bw(self.data_bw, self.greens_bw, q_grid)
                misfit += res_bw
        except:
            pass
        try:
            if self.data_sw is not None and self.greens_sw is not None:
                res_sw = self.misfit_sw(self.data_sw, self.greens_sw, q_grid)
                misfit += res_sw
        except:
            pass

        # misfit scaling
        misfit /= self.scaling

        Uq = misfit / (2.0 * np.power(self.sigma_d, 2))
        # gradient: dU/dqi = (y(x+dx) - y(x))/dx
        self.dU_dqi = np.zeros(self.Nq)
        for i in range(self.Nq):
            self.dU_dqi[i] = (Uq[i + 1][0] - Uq[0][0]) / self.delta_q0[i]

        self.dU_dqi[3] /= 100.  # magnitude

        return Uq[0][0]


    def set_q(self, q):
        '''
        Set position: q.
        q = [strike, dip, rake, magnitude, co-latitude, lune-longitude] in degree and Mw
        '''

        if len(q) < self.Nq:
            ValueError("q should contain [strike, dip, rake, magnitude, co-latitude, lune-longitude].")

        self.q0 = np.zeros(self.Nq)
        self.q0[0] = wrap_strike(q[0])
        self.q0[1] = wrap_dip(q[1])
        self.q0[2] = wrap_rake(q[2])
        self.q0[3] = q[3]
        self.q0[4] = wrap_colatitude(q[4])
        self.q0[5] = wrap_lunelongitude(q[5])

        if not self.b_initial_q0:
            self.raw_q0 = self.q0.copy()

        self.b_initial_q0 = True
        self.Uq(self.q0)



# -------------------------------------------------------------------
# Tools and utilities
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import numpy as np


def wrap_strike(angle_in_deg):
    '''Wrap strike angle to (0, 360)'''
    if angle_in_deg < 0:
        return np.mod(angle_in_deg+360., 360.)
    if angle_in_deg > 360.:
        return np.mod(angle_in_deg, 360.)
    return angle_in_deg


def wrap_dip(angle_in_deg):
    '''Wrap dip angle to (0, 90)'''
    if angle_in_deg > 90:
        return 90. - np.mod(angle_in_deg, 90)
    if angle_in_deg < 0:
        return np.fabs(angle_in_deg)
    return angle_in_deg


def wrap_rake(angle_in_deg):
    '''Wraps slip angle to (-90, 90)'''
    if angle_in_deg > 90:
        return 90. - np.mod(angle_in_deg, 90)
    if angle_in_deg < -90:
        return -90. - np.mod(angle_in_deg, -90)
    return angle_in_deg


def wrap_colatitude(angle_in_deg):
    '''Wrap colatitude to (0, 180)'''
    if angle_in_deg > 180:
        return 180. - np.mod(angle_in_deg, 180)
    if angle_in_deg < 0:
        return np.fabs(angle_in_deg)
    return angle_in_deg


def wrap_lunelongitude(angle_in_deg):
    '''Wrap colatitude to (-30, 30)'''

    if angle_in_deg > 30:
        return 30. - np.mod(angle_in_deg, 30)
    if angle_in_deg < -30:
        return -30. - np.mod(angle_in_deg, -30)
    return angle_in_deg




# -------------------------------------------------------------------
# Functions to determine the solution
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import pandas as pd
import numpy as np


def find_sol(df, type='misfit'):
    '''Determine the solution based on either:
    (1) The lowest Uq that is associated with waveform misfit (by default), or
    (2) Sample density.
    '''
    if type == 'misfit':
        return find_sol_misfit(df)
    else:
        return find_sol_sample_density(df)


def find_sol_sample_density(df):
    '''Find the solution based on sample density. '''

    keys = df.keys()
    sol = np.zeros(len(keys))
    bins = 25
    for i, k in enumerate(keys):
        hist, edges = np.histogram(df[k].values[:], bins=bins)
        sol[i] = edges[np.where(hist == np.max(hist))][0]
    return sol


def find_sol_misfit(df):
    '''Find the solution based on the lowest Uq (associated with waveform misfit) '''
    keys = df.keys()
    _sol = df[df['misfit'] == df['misfit'].min()]
    return [_sol[k].values[0] for k in keys]




# -------------------------------------------------------------------
# Moment tensor inversion using Hamiltonian Monte Carlo sampling
# The base class
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import os.path
import json
import numpy as np
import pickle
import time

class DHMCBase():
    '''The base class of HMC sampling'''

    def __init__(self, saving_dir,
                 b_save_cache=True, n_step_cache=10,
                 verbose=False):

        ##############################################################
        # Following parameters associated with Nq must be implemented by subclass.
        self.Nq = 2
        # Use identity matrix as the mass matrix by default.
        self.massinv    = np.eye(self.Nq)
        self.delta_q0   = 0.1 * np.ones(self.Nq)
        self.q0         = np.zeros(self.Nq)
        self.raw_q0     = np.zeros(self.Nq)
        self.dU_dqi     = np.zeros(self.Nq)
        # solution at lowest Uq (misfit)
        self.sol_misfit = np.zeros(self.Nq)
        ##############################################################

        # default parameters for sampling
        # constant encoding the uncertainties in observations
        self.sigma_d = 0.1
        self.sigma_q = np.inf

        # The saving directory
        self.saving_dir = saving_dir
        if not os.path.exists(self.saving_dir):
            os.makedirs(self.saving_dir)

        # Saving cache.
        self.b_save_cache = b_save_cache
        self.n_step_cache = n_step_cache

        self.b_init_epsilon = False
        self.b_init_n_step = False
        self.b_initial_q0 = False

        # Terminate if exceed the MAX_SAMPLE.
        self.MAX_SAMPLE = 100000
        self.verbose = verbose


    def _check(self):
        '''Check the required parameters.'''
        if (False in [self.b_init_n_step, self.b_init_epsilon, self.b_initial_q0]):
            return False
        else:
            return True

    def create_p(self):
        '''Using probability density function to create P. '''
        return np.random.normal(loc=0, scale=1.0, size=self.Nq)

    def Kp(self, p):
        '''Compute Kp = 1/2P'MP, M:mass matrix. '''
        return 0.5 * p.T @ (self.massinv @ p)

    def Uq(self, q):
        raise NotImplementedError("Must be implemented by subclass")

    def set_q(self, q):
        raise NotImplementedError("Must be implemented by subclass")

    def dUdq(self, q):
        raise NotImplementedError("Must be implemented by subclass")

    def set_epsilon(self, min, max):
        '''Set function to generate epsilon in sampling.'''
        self.min_epsilon = min
        self.max_epsilon = max
        self.b_init_epsilon = True

    def get_epsilon(self):
        '''Get epsilon'''
        if not self.b_init_epsilon:
            raise ValueError("Uninitialized epsilon !!!")
        else:
            return np.random.uniform(self.min_epsilon, self.max_epsilon)

    def set_n_step(self, min, max=None):
        '''Set nstep '''
        self.min_n_step = np.uint(min)
        self.max_n_step = np.uint(max)
        self.b_init_n_step = True

    def get_n_step(self):
        '''Get n_Step'''
        if not self.b_init_n_step:
            raise ValueError("Uninitialized n_step !!!")
        else:
            return int(np.random.uniform(self.min_n_step, self.max_n_step))

    def _find_solution(self, samples, misfit):
        '''Determine the solution based on the lowest misfit. '''
        self.sol_misfit = np.array(samples[np.where(misfit == np.min(misfit))[0][0]])

    def _get_solution(self):
        '''Return the solution.'''
        return self.sol_misfit

    def hmc(self):
        '''Hamiltonian Monte Carlo Sampling.'''
        if not self.b_initial_q0:
            ValueError("Uninitialized q0!")

        # Set the q(t=0) and p(t=0).
        q = self.q0.copy()
        current_q = self.q0.copy()
        p = self.create_p()
        current_p = p.copy()

        # set step size and number of step
        epsilon = self.get_epsilon()
        n_step = self.get_n_step()

        for i in range(n_step):
            # p(t+epsilon/2)
            p = p - (epsilon / 2.0) * self.dU_dqi
            # q(t+epsilon)
            q = q + epsilon * self.massinv @ p
            self.set_q(q)
            p = p - (epsilon / 2.0) * self.dU_dqi

        current_Uq = self.Uq(current_q)
        current_Kp = self.Kp(current_p)
        proposed_Uq = self.Uq(q)
        proposed_Kp = self.Kp(p)
        denergy = current_Uq - proposed_Uq + current_Kp - proposed_Kp

        # accept the sample with probability
        if np.isnan(denergy): denergy = -np.inf
        prob_accept = min(0, denergy)
        alpha = np.log(np.random.rand())
        # accept q(t=n_step), use it as the new q(t=0)
        if alpha < prob_accept:
            self.set_q(q)
            if self.verbose:
                print("! Accept q={}".format(np.round(self.q0, 2)))
            return self.q0, True, proposed_Uq
        # reject q(t=n_step), return to q(t=0)
        else:
            if self.verbose:
                print("Reject, q={}".format(np.round(q, 2)))
            self.set_q(current_q)
            return current_q, False, proposed_Uq


    def sampling(self, n_sample, task_id='hmc01'):
        '''Sampling process'''
        if not self.b_initial_q0:
            raise ValueError("q0 must be initialized!")

        if not self._check():
            raise ValueError("Initialized parameters: n_step, epsilon, ...")

        t0 = time.time()
        # data container
        samples = []
        misfit = []
        _tmp_n_sample = 0
        while (len(samples) < n_sample and _tmp_n_sample < self.MAX_SAMPLE):
            _tmp_n_sample += 1
            try:
                q, accept, U = self.hmc()
                if accept:
                    samples.append(q)
                    misfit.append(U)
            except:
                # reset
                self.set_q(self.raw_q0.copy())

            # save cache
            if self.b_save_cache:
                if len(samples) != 0 and np.mod(len(samples), self.n_step_cache) == 0:
                    self.save_samples('%s_cache' % task_id, samples, misfit)

        # save all samples
        self.save_samples('%s_samples' % task_id, samples, misfit)

        # determine solution based on waveform misfit
        self._find_solution(samples, misfit)

        # gather the information of HMC sampling
        dt = '%d' % (time.time() - t0)
        ar = '%.2f' % (len(samples) / _tmp_n_sample)    # accept rate
        info_dict = {
            'Task_id': task_id,
            'time_elapse_in_s': dt,
            'N_sample': len(samples),
            'Accept_rate': ar,
            'Nq': self.Nq,
            'Sigma_d': self.sigma_d,
            "q0": str(self.raw_q0),
            "sol_misfit": str(samples[np.where(misfit == np.min(misfit))[0][0]]),
        }
        self.save_info('%s_info' % task_id, info_dict)


    def save_samples(self, file_name, samples, misfit, fmt='pkl'):
        ''' Save all accepted samples to a PKL file. '''

        if not os.path.exists(self.saving_dir):
            os.makedirs(self.saving_dir)

        file_path = os.path.join(self.saving_dir, "%s_N%d.%s"%(file_name, len(samples), fmt))
        with open(file_path, 'wb') as f:
            pickle.dump(samples, f)
            pickle.dump(misfit, f)

    def save_info(self, file_name, info_dict, fmt='json'):
        '''Save the information into a json file'''

        if not os.path.exists(self.saving_dir):
            os.makedirs(self.saving_dir)

        file_path = os.path.join(self.saving_dir, "%s.%s"%(file_name, fmt))
        with open(file_path, 'w', encoding='utf-8') as file:
            json.dump(info_dict, file, indent=4)



import setuptools

setuptools.setup(
    name="seishmc",
    version="0.0.1",
    author="Liang Ding",
    author_email="myliang.ding@mail.utoronto.ca",
    description="Full moment tensor inversion using Hamiltonian Monte Carlo (HMC) sampling",
    long_description="Full moment tensor inversion using Hamiltonian Monte Carlo (HMC) sampling",
    long_description_content_type="text/markdown",
    url="https://github.com/Liang-Ding/seishmc",
    project_urls={
        "Bug Tracker": "https://github.com/Liang-Ding/seishmc/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=[
        "seismology"
    ],
    # package_dir={"": "seishmc"},
    python_requires='>=3.7.0',
    install_requires=[
        "numpy",
        "pandas",
        "seaborn>=0.11.2",
    ],
    packages=setuptools.find_packages(),
)


# -------------------------------------------------------------------
# Moment tensor inversion using Hamiltonian Monte Carlo sampling
# Double-couple (DC) solutions
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

from seishmc.utils import wrap_strike, wrap_rake, wrap_dip
from seishmc.DHMC.base import DHMCBase

from mtuq.grid import UnstructuredGrid
from mtuq.util.math import to_mij, to_rho
from mtuq.grid.moment_tensor import to_mt
from mtuq.event import MomentTensor
import numpy as np

class DHMC_DC(DHMCBase):
    ''' Utilizing the Hamiltonian Monte Carlo sampling method to sample the double-couple solutions.'''

    def __init__(self, misfit_bw, data_bw, greens_bw,
                 misfit_sw, data_sw, greens_sw,
                 saving_dir, b_save_cache=True, n_step_cache=100,
                 verbose=False):

        super(DHMC_DC, self).__init__(saving_dir, b_save_cache, n_step_cache, verbose)

        # double-couple solutions
        # [strike, dip, rake, magnitude]
        self.Nq = 4
        # mass matrix
        self.massinv = np.eye(self.Nq)
        self.massinv[3, 3] = 1.0 / 1000.    # magnitude
        # delta to calculate derivatives.
        self.delta_q0 = 0.2 * np.ones(self.Nq)
        self.delta_q0[3] = 0.02  # magnitude
        self.q0 = np.zeros(self.Nq)
        self.raw_q0 = np.zeros(self.Nq)
        self.dU_dqi = np.zeros(self.Nq)
        self.sol_misfit = np.zeros(self.Nq)

        # body wave
        self.misfit_bw = misfit_bw
        self.data_bw = data_bw
        self.greens_bw = greens_bw
        # surface wave
        self.misfit_sw = misfit_sw
        self.data_sw = data_sw
        self.greens_sw = greens_sw

        self.calc_scaling_factor()


    def get_solution(self):
        '''Get solution'''

        sol = self._get_solution()
        kappa   = sol[0]  # strike
        h       = np.cos(np.deg2rad(sol[1]))  # dip
        sigma   = sol[2]  # rake
        rho     = to_rho(sol[3])  # magnitude: Mw
        v       = 0.
        w       = 0.
        mij     = to_mij(rho, v, w, kappa, sigma, h)

        # lune dict
        self.lune_dict = dict({'rho': rho,
                          'v': v,
                          'w': w,
                          'kappa': kappa,
                          'sigma': sigma,
                          'h': h, })
        self.best_mt = MomentTensor(mij)
        return self.best_mt, self.lune_dict


    def calc_scaling_factor(self):
        '''Compute misfit scaling factor based on the data.'''
        self.scaling = 0.0

        # body wave
        if self.data_bw is not None and self.greens_bw is not None:
            for _st in self.data_bw:
                for _tr in _st:
                    self.scaling += np.dot(_tr.data, _tr.data)

        # surface wave
        if self.data_sw is not None and self.greens_sw is not None:
            for _st in self.data_sw:
                for _tr in _st:
                    self.scaling += np.dot(_tr.data, _tr.data)

        if self.scaling == 0.0:
            self.scaling = 1.0


    def Uq(self, q):
        '''Compute Uq and gradient'''

        npts = self.Nq + 1
        v = np.zeros(npts)
        w = np.zeros(npts)
        kappa = q[0] * np.ones(npts)                    # strike
        h = np.cos(np.deg2rad(q[1])) * np.ones(npts)    # dip
        sigma = q[2] * np.ones(npts)                    # rake
        rho = to_rho(q[3]) * np.ones(npts)              # magnitude: Mw

        # x+dx
        kappa[1] += self.delta_q0[0]
        h[2] = np.cos(np.deg2rad(q[1]+self.delta_q0[1]))
        sigma[3] += self.delta_q0[2]
        rho[4] = to_rho(q[3]+self.delta_q0[3])

        q_grid = UnstructuredGrid(
            dims=('rho', 'v', 'w', 'kappa', 'sigma', 'h'),
            coords=(rho, v, w, kappa, sigma, h),
            callback=to_mt)

        misfit = np.zeros([npts, 1])

        try:
            if self.data_bw is not None and self.greens_bw is not None:
                res_bw = self.misfit_bw(self.data_bw, self.greens_bw, q_grid)
                misfit += res_bw
        except:
            pass
        try:
            if self.data_sw is not None and self.greens_sw is not None:
                res_sw = self.misfit_sw(self.data_sw, self.greens_sw, q_grid)
                misfit += res_sw
        except:
            pass

        # misfit scaling
        misfit /= self.scaling

        Uq = misfit / (2.0 * np.power(self.sigma_d, 2))
        # gradient: dU/dqi = (y(x+dx) - y(x))/dx
        self.dU_dqi = np.zeros(self.Nq)
        for i in range(self.Nq):
            self.dU_dqi[i] = (Uq[i+1][0] - Uq[0][0]) / self.delta_q0[i]

        self.dU_dqi[3] /= 100.   # magnitude
        return Uq[0][0]


    def set_q(self, q):
        '''
        Set position: q.
        q = [strike, dip, rake, magnitude] in degree and Mw
        '''

        if len(q) < self.Nq:
            ValueError("q should contain [strike, dip, rake, magnitude].")

        self.q0 = np.zeros(self.Nq)
        self.q0[0] = wrap_strike(q[0])
        self.q0[1] = wrap_dip(q[1])
        self.q0[2] = wrap_rake(q[2])
        self.q0[3] = q[3]

        if not self.b_initial_q0:
            self.raw_q0 = self.q0.copy()

        self.b_initial_q0 = True
        self.Uq(self.q0)

